unit Undo;
{an unit for undoing editing operations, like moving, adding, pasting components}
interface

uses
  SysUtils, WinTypes, Dialogs, Classes, Skemadef, Utildef, Diagram, Drawmenu;

type
{class for saving the (editing) operations with parameters}
  TpUndoOp = ^TUndoOp;        {define pointer type first !}
  TUndoOp = class(TObject)
  {public procedures}
    procedure PutpSet(pSet: TpCompSet);
    procedure AddForUndo(pSet: TpCompSet; untype: TUndoType; forUndo: boolean);
    procedure FreeLists;
    procedure Initialize;
    function  UndoOperation(redo: boolean): boolean;
    function  GetUntype: TUndoType;
  public
    {public variables}
    ii_u       : integer;     {index}
    text_u     : string;      {description of the operation to be undone}
    untype_u   : TUndoType;   {type of operation do be undone}
  private
    {private variables (and procedures)}
    pSet_u     : TpCompSet;   {pointer to the common}
    pComp_u    : TpComp;      {pointer to a single component}
    CompList_u : TList;       {pointer to a list of components, for undoing move, paste}
    GraphList_u: TList;       {list for graphical symbols ("shapes")}
    ix_u,iy_u  : integer;     {distance moved}
end;



implementation

procedure TUndoOp.PutpSet(pSet:TpCompSet);
{put the pointer to the "common" to the object variable}
begin
  pSet_u := pSet;
end;

procedure TUndoOp.initialize;
begin
  CompList_u := nil;
  GraphList_u := nil;
end;

procedure TUndoOp.FreeLists;
begin
  CompList_u.free;
  CompList_u := nil;    {must set to nil, if UndoOp is reused, not deleted}
  GraphList_u.free;
  GraphList_u := nil;
end;

procedure TUndoOp.AddForUndo(pSet: TpCompSet; untype: TUndoType; forUndo: boolean);
{create an object that saves the data for the undo operation
 pSet     pointer to the common
 untype   the operation to be undone
 forUndo  true, if saved for undo, false if saved for redo}
{called in SaveForUndo and SaveForRedo in Skemadef.pas}
const
  udcount : integer = 0;
var
  ii    : integer;
  jdata : integer;  {counts lines}
  line  : string;   {one line of data}
  nc    : integer;  {number of components, so far}
  nl    : integer;
  pLast : TpComp;   {pointer to the last added component}
begin
  {Note: object created after an operation to be saved. The creation done in Skemadef.pas}
  pSet_u := pSet;            {save the pointer to the "common"}
  untype_u := untype;        {the type of the operation saved for undo}
  {the parameters are interpreted according to the operation}
ii_u := udcount;
inc(udcount);
  if (untype = MoveGroup) then begin
    ix_u := pSet_u^.move_ddx;           {movement in x-direction}
    iy_u := pSet_u^.move_ddy;
    {copy the list of components}
    CompList_u := TList.create;         {first, create the list}
    CompList_u.clear;                   {just to be sure}
    if (pSet^.CompList.count > 0) then begin
      for ii := 0 to pSet^.CompList.count-1 do CompList_u.add(pSet^.CompList[ii]);
    end;
    {graphical elements moved (if any)}
    GraphList_u := TList.create;

    DrawMenuForm.CopyInsideToList(GraphList_u);  {copy pointers to moved graphical elements}
    text_u := 'move group';
  end else if (untype_u = MoveComp) then begin {moving one component}
    pComp_u := TpComp(pSet_u^.Appl_temp);      {pointer in appl_temp}
    ix_u := pSet_u^.move_ddx;                  {movement in x-direction}
    iy_u := pSet_u^.move_ddy;
    text_u := 'move one component';
  end else if (untype_u = MoveAll) then begin  {moving all with the arrow keys}
    ix_u := pSet_u^.move_ddx;                  {save only the movement}
    iy_u := pSet_u^.move_ddy;
    text_u := 'move all';
  end else if (untype_u = AddComp) then begin  {add a single component}
    nc := pSet_u^.pCompList.count;
    if (nc > 0) then begin                     {might = 0, when adding dummy comp}
      pLast := pSet_u^.pCompList[nc-1];
      pComp_u := pLast;                        {save only a pointer to the component}
      if (forUndo) then
        text_u := 'add a component'
      else
        text_u := 'delete a component';        {redoing = undo-undo}
    end;
  end else if (untype_u = AddGroup) then begin {add several components}
    CompList_u := TList.create;                {first, create the list}
    CompList_u.clear;                          {just to be sure}
    if (pSet^.CompList.count > 0) then begin   {copy from the list of created components}
      for ii := 0 to pSet^.CompList.count-1 do CompList_u.add(pSet^.CompList[ii]);
    end;
    if (forUndo) then
      text_u := 'add components'
    else
      text_u := 'delete';
  end else if (untype_u = Delete) then begin   {delete one or more components}
    CompList_u := TList.create;                {create list for saving pointers to comp.}
    CompList_u.clear;                          {just to be sure}
    nl := pSet_u^.CompList.count;
    jdata := 0;
    for ii := 0 to nl-1 do begin
      CompList_u.add(pSet_u^.CompList[ii]);
    end;
    if (forUndo) then
      text_u := 'delete'
    else
      text_u := 'add component(s)';
  end else begin

  end;
//  pSet^.CompList.clear;  {list not needed any more, may confuse, if not empty..}
  {...when undo delete + not in use => cannot change the InUse-state,
      or takes wrong InUse-state because "not CompInList"-check in
      TComp.Mouse_down, in Skemadef.pas 28.10.09
      the "not CompInList-check removed 31.10.09 (any problems?)}
  {...but paste does not work, if list cleared ??}
end;

function TUndoOp.UndoOperation(redo: boolean): boolean;
{carry out the actual undo-operation
 redo   true, if redo, not undo
 return "true", if the undo-operation was carried out}
var
  answer  : word;     {answer to a warning}
  dxy     : TPoint;   {dx,dy combined}
  errc    : integer;  {error flag}
  highli  : boolean;
  ii      : integer;  {index}
  jcomp   : integer;  {counts components}
  ll      : longint;  {temporary}
  messtxt : string;   {text to be (optionally) displayed}
  moved   : boolean;  {components really moved}
  nc      : integer;  {number of components}
  pComp   : TpComp;   {pointer to a component}
  pDiagr  : TpDiagramForm;  {pointer to the zoomed-out diagram}
  pForm   : TpForm;   {pointer to the main form}
  pMod    : TpElComp; {pointer to the model component}
  undoadd : boolean;  {undo the addition of components}
begin
  {initially, hide the markings, for simplicity}
  if (pSet_u^.pMovSch <> nil) then begin
    pSet_u^.pMovSch^.enabled := false;
    pSet_u^.pMovSch^.visible := false;
    pSet_u^.pMovSch^.SendToBack;
    pSet_u^.pMovSch^.picture := nil;
  end;
  {hide image for ctrl-selected comp.}
  if pSet_u^.pCtrl_img^.visible then pSet_u^.HideCtrlImage;
  {next, the actual undo of operations}
  if (untype_u = MoveGroup) then begin  {undo a move of a group of components}
    pSet_u^.CompListMove(CompList_u, -ix_u, -iy_u, not redo, highli, moved);
    {move also the graphical elements (if any) backwards one step}
     DrawMenuForm.MoveInList(GraphList_u,-ix_u,-iy_u);
     DrawMenuForm.CopyListToInside(GraphList_u);       {copy for redo}
    {reselect the components}  {is this needed ??}
    if (CompList_u.count > 0) then begin
      pSet_u^.PrevList.clear;
      for ii := 0 to CompList_u.count-1 do begin
        pSet_u^.PrevList.add(CompList_u[ii]);
      end;
    end;
    {copy the list of components for undoing/redoing }
    if (pSet_u^.un_doing) or (pSet_u^.re_doing) then begin  {is this needed?}
      pSet_u^.CompList.clear;
      if (CompList_u.count > 0) then begin
        for ii := 0 to CompList_u.count-1 do begin
          pSet_u^.CompList.add(CompList_u[ii]);
        end;
      end;
    end;
    UndoOperation := true;
    {save for redo, if this is undo}
    if (pSet_u^.un_doing) or (pSet_u^.re_doing) then begin
      pSet_u^.move_ddx := -ix_u;       {must copy to _ddx, because AddForUndo uses it...}
      pSet_u^.move_ddy := -iy_u;       {..and AddForUndo is used also for redo}
      if (pSet_u^.un_doing) then pSet_u^.SaveForRedo(MoveGroup)
      else pSet_u^.SaveForUndo(MoveGroup);
    end;
    messtxt := 'undo move components';
  end else if (untype_u = MoveComp) then begin {moving a single component}
    pComp_u^.IncrCoord(-ix_u,-iy_u);
    UndoOperation := true;
    messtxt := 'undo move one component';
    {save for redo, if this is undo, and save for undo, if this is redo}
    if (pSet_u^.un_doing) or (pSet_u^.re_doing) then begin
      pSet_u^.Appl_temp := longint(pComp_u); {..must copy, se next..}
      pSet_u^.move_ddx := -ix_u;       {must copy to _ddx, because AddForUndo uses it...}
      pSet_u^.move_ddy := -iy_u;       {..and AddForUndo is used also for redo}
      if  (pSet_u^.un_doing) then pSet_u^.SaveForRedo(MoveComp)
      else pSet_u^.SaveForUndo(MoveComp);
    end;
  end else if (untype_u = MoveAll) then begin  {move all with the arrow keys}
    pComp := nil;
    errc := 0;
    jcomp := 0;
    while (errc = 0) do begin                    {scan all components}
      pSet_u^.NextComp(pComp, errc,jcomp,false); {pointer to next component}
      if (errc = 0) then begin                   {no errors, no more components}
        pComp^.IncrCoord(-ix_u,-iy_u);
      end;
    end;
    {move also the graphical elements (if any) backwards one step}
    DrawMenuForm.MoveAll(-ix_u,-iy_u);
    {update also the marker position in the zoomed diagram }
    pDiagr := TpDiagramForm(pSet_u^.ipDiagram);
    dxy.x := -ix_u;  {pack in one variable}
    dxy.y := -iy_u;
    ll := longint(@dxy);
    pDiagr^.MarkerPosition(UndoMoveAll,ll);
    UndoOperation := true;
    messtxt := 'undo move all components';
    {save for redo, if this is undo, and save for undo, if this is redo}
    if (pSet_u^.un_doing) or (pSet_u^.re_doing) then begin
      pSet_u^.move_ddx := -ix_u;                  {save only the movement}
      pSet_u^.move_ddy := -iy_u;
      if  (pSet_u^.un_doing) then pSet_u^.SaveForRedo(MoveAll)
      else pSet_u^.SaveForUndo(MoveAll);
    end;
  end else if (untype_u = AddComp) then begin  {undo adding a single component}
    pComp_u^.HideComp;
    UndoOperation := true;
    messtxt := 'undo add one component';
    {save for redo, if this is undo, and save for undo, if this is redo}
    if (pSet_u^.un_doing) or (pSet_u^.re_doing) then begin
      pSet_u^.CompList.clear;
      pSet_u^.CompList.add(pComp_u);
      if (pSet_u^.un_doing) then begin
        pSet_u^.SaveForRedo(Delete);
      end else begin
        pSet_u^.SaveForUndo(Delete);
      end;
    end;
  end else if (untype_u = AddGroup) then begin {undo adding several components}
    nc := CompList_u.count;                    {number of added components}
    {give a warning, if clearing all components}
    if (nc = pSet_u^.pCompList.count) then begin
      answer := MessageDlg('Delete the whole diagram?',
      mtWarning, [mbYes, mbCancel], 0);        {else ask for confirmation}
      undoadd :=  (answer = mrYes);            {undo accepted}
    end else begin                             {not the last components, undo}
      undoadd := true;
    end;
    {undo the addition, if allowed}
    if (undoadd) then begin
      pSet_u^.CompList.clear;
      if (nc > 0) then begin
        for ii := 0 to nc-1 do begin
          pComp := CompList_u[ii];
          pSet_u^.CompList.add(pComp);        {add to list for redo or undo}
          if (pComp <> nil) then begin
            pComp^.HideComp;
          end;
        end;
      end;
      UndoOperation := true;
      {save for redo, if this is undo, and save for undo, if this is redo}
      if (pSet_u^.un_doing) then begin
        pSet_u^.SaveForRedo(Delete);
      end else begin
        pSet_u^.SaveForUndo(Delete);
      end;
    end else begin
      UndoOperation := false;
    end;
  end else if (untype_u = Delete) then begin  {undeleting one or more components}
    {first, save for redo, if this is undo, and save for undo, if this is redo}
    if (pSet_u^.un_doing) or (pSet_u^.re_doing) then begin
      pSet_u^.CompList.clear;
      nc := CompList_u.count;
      if (nc > 0) then begin
        for ii := 0 to nc-1 do begin
          pComp := CompList_u[ii];
          pSet_u^.CompList.add(pComp{CompList_u[ii]});
//          {if component not in use, change color to light gray, if allowed}
//          pComp^.GetipMod(ll);               {pointer to model comp., as integer}
//          pMod := TpElComp(ll);              {pointer to the model component}
//          if (pMod^.tag and InUseMask <> 0) then  begin
//            CompGreyIf(pComp);
//          end;
        end;
        if (pSet_u^.un_doing) then pSet_u^.SaveForRedo(AddGroup)
        else pSet_u^.SaveForUndo(AddGroup);
      end;
    end;
    {then, do the actual undo (or redo)}
    pSet_u^.Undelete(CompList_u);             {note: undelete clears the list}
    UndoOperation := true;
    messtxt := 'undo delete component(s)';
  end else begin
    UndoOperation := false;
  end;
  {flag that the diagram has (probably) changed}
  pSet_u^.IsChanged := true;
end;

function  TUndoOp.GetUntype: TUndoType;
begin
  GetUntype := untype_u;
end;

end.
